from cached_property import cached_property
from collections import OrderedDict
from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import JSON
from onegov.file import AssociatedFiles
from onegov.file import File
from onegov.file.attachments import extract_pdf_info
from onegov.swissvotes import _
from onegov.swissvotes.models.actor import Actor
from onegov.swissvotes.models.localized_file import LocalizedFile
from onegov.swissvotes.models.policy_area import PolicyArea
from onegov.swissvotes.models.region import Region
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import func
from sqlalchemy import Integer
from sqlalchemy import Numeric
from sqlalchemy import Text
from sqlalchemy_utils import observes
from sqlalchemy.dialects.postgresql import INT4RANGE
from sqlalchemy.dialects.postgresql import TSVECTOR
from sqlalchemy.orm import deferred
from urllib.parse import urlparse
from urllib.parse import urlunparse


class SwissVoteFile(File):
    """ An attachment to a vote. """

    __mapper_args__ = {'polymorphic_identity': 'swissvote'}


class encoded_property(object):
    """ A shorthand property to return the label of an encoded value. Requires
    the instance the have a `codes`-lookup function. Assumes that the value
    to lookup has the same name as the property prefixed with an underline.

    Example:

        class MyClass(object):
            _value = 0
            value = encoded_property()

            def codes(self, attributes):
                return {0: 'None', 1: 'One'}

    """

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner):
        value = getattr(instance, f'_{self.name}')
        return instance.codes(self.name).get(value)


class SwissVote(Base, TimestampMixin, AssociatedFiles):

    """ A single vote as defined by the code book.

    There are a lot of columns:
    - Some general, ready to be used attributes (bfs_number, ...)
    - Encoded attributes, where the raw integer value is stored prefixed with
      an underline and the attribute returns a translatable label by using the
      ``codes`` function, e.g.  ``_legal_form``, ``legal_form`` and
      ``codes(' _legal_form')``.
    - Descriptors, easily accessible by using ``policy_areas``.
    - A lot of lazy loaded, cantonal results only used when importing/exporting
      the dataset.
    - Recommendations from different parties and assocciations. Internally
      stored as JSON and easily accessible and group by slogan with
      ``recommendations_parties``, ``recommendations_divergent_parties`` and
      ``recommendations_associations``.
    - Different localized attachments, some of them indexed for full text
      search.

    """

    __tablename__ = 'swissvotes'

    ORGANIZATION_NO_LONGER_EXISTS = 9999

    @staticmethod
    def codes(attribute):
        """ Returns the codes for the given attribute as defined in the code
        book.

        """
        if attribute == 'legal_form':
            return OrderedDict((
                (1, _("Mandatory referendum")),
                (2, _("Optional referendum")),
                (3, _("Popular initiative")),
                (4, _("Direct counter-proposal")),
            ))
        if attribute == 'result_cantons_accepted':
            return OrderedDict((
                (0, _("Rejected")),
                (1, _("Accepted")),
                (3, _("Majority of the cantons not necessary")),
            ))
        if attribute == 'result' or attribute.endswith('_accepted'):
            return OrderedDict((
                (0, _("Rejected")),
                (1, _("Accepted")),
            ))
        if attribute == 'department_in_charge':
            return OrderedDict((
                (1, _("Federal Department of Foreign Affairs (FDFA)")),
                (2, _("Federal Department of Home Affairs (FDHA)")),
                (3, _("Federal Department of Justice and Police (FDJP)")),
                (4, _("Federal Department of Defence, Civil Protection and "
                      "Sport (DDPS)")),
                (5, _("Federal Department of Finance (FDF)")),
                (6, _("Federal Department of Economic Affairs, Education and "
                      "Research (EAER)")),
                (7, _("Federal Department of the Environment, Transport, "
                      "Energy and Communications (DETEC)")),
                (8, _("Federal Chancellery (FCh)")),
            ))
        if (attribute == 'position_federal_council'
                or attribute == 'position_national_council'
                or attribute == 'position_council_of_states'):
            return OrderedDict((
                (1, _("Accepting")),
                (2, _("Rejecting")),
                (3, _("None"))
            ))
        if attribute == 'position_parliament':
            return OrderedDict((
                (1, _("Accepting")),
                (2, _("Rejecting")),
            ))
        if attribute == 'recommendation':
            # Added ordering how it should be displayed in strengths table
            return OrderedDict((
                (1, _("Yea")),
                (2, _("Nay")),
                (4, _("Empty")),
                (5, _("Free vote")),
                (3, _("None")),
                (66, _("Neutral")),
                (9999, _("Organization no longer exists")),
                (None, _("unknown"))
            ))

        raise RuntimeError(f"No codes available for '{attribute}'")

    id = Column(Integer, nullable=False, primary_key=True)

    # Formal description
    bfs_number = Column(Numeric(8, 2), nullable=False)
    date = Column(Date, nullable=False)
    legislation_number = Column(Integer, nullable=False)
    legislation_decade = Column(INT4RANGE, nullable=False)
    title_de = Column(Text, nullable=False)
    title_fr = Column(Text, nullable=False)
    short_title_de = Column(Text, nullable=False)
    short_title_fr = Column(Text, nullable=False)
    keyword = Column(Text)
    votes_on_same_day = Column(Integer, nullable=False)
    _legal_form = Column('legal_form', Integer, nullable=False)
    legal_form = encoded_property()
    initiator = Column(Text)
    anneepolitique = Column(Text)
    bfs_map_de = Column(Text)
    bfs_map_fr = Column(Text)

    @property
    def title(self):
        if self.session_manager.current_locale == 'fr_CH':
            return self.title_fr
        else:
            return self.title_de

    @property
    def short_title(self):
        if self.session_manager.current_locale == 'fr_CH':
            return self.short_title_fr
        else:
            return self.short_title_de

    def bfs_map(self, locale):
        """ Returns the link to the BFS map for the given locale. """

        return self.bfs_map_fr if locale == 'fr_CH' else self.bfs_map_de

    def bfs_map_host(self, locale):
        """ Returns the Host of the BFS Map link for CSP. """

        try:
            return urlunparse(
                list(urlparse(self.bfs_map(locale))[:2]) + ['', '', '', '']
            )
        except ValueError:
            pass

    # Descriptor
    descriptor_1_level_1 = Column(Numeric(8, 4))
    descriptor_1_level_2 = Column(Numeric(8, 4))
    descriptor_1_level_3 = Column(Numeric(8, 4))
    descriptor_2_level_1 = Column(Numeric(8, 4))
    descriptor_2_level_2 = Column(Numeric(8, 4))
    descriptor_2_level_3 = Column(Numeric(8, 4))
    descriptor_3_level_1 = Column(Numeric(8, 4))
    descriptor_3_level_2 = Column(Numeric(8, 4))
    descriptor_3_level_3 = Column(Numeric(8, 4))

    @cached_property
    def policy_areas(self):
        """ Returns the policy areas / descriptors of the vote. """

        def get_level(number, level):
            value = getattr(self, f'descriptor_{number}_level_{level}')
            if value is not None:
                return PolicyArea(value, level)

        result = []
        for number in (1, 2, 3):
            for level in (3, 2, 1):
                area = get_level(number, level)
                if area:
                    result.append(area)
                    break
        return result

    # Result
    _result = Column('result', Integer)
    result = encoded_property()
    result_eligible_voters = deferred(Column(Integer), group='dataset')
    result_votes_empty = deferred(Column(Integer), group='dataset')
    result_votes_invalid = deferred(Column(Integer), group='dataset')
    result_votes_valid = deferred(Column(Integer), group='dataset')
    result_votes_total = deferred(Column(Integer), group='dataset')
    result_turnout = Column(Numeric(13, 10))

    _result_people_accepted = Column('result_people_accepted', Integer)
    result_people_accepted = encoded_property()
    result_people_yeas = Column(Integer)
    result_people_nays = Column(Integer)
    result_people_yeas_p = Column(Numeric(13, 10))

    _result_cantons_accepted = Column('result_cantons_accepted', Integer)
    result_cantons_accepted = encoded_property()
    result_cantons_yeas = Column(Numeric(3, 1))
    result_cantons_nays = Column(Numeric(3, 1))
    result_cantons_yeas_p = Column(Numeric(13, 10))

    _result_ag_accepted = Column('result_ag_accepted', Integer)
    result_ag_accepted = encoded_property()
    result_ag_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ag_votes_valid = deferred(Column(Integer), group='dataset')
    result_ag_votes_total = deferred(Column(Integer), group='dataset')
    result_ag_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ag_yeas = deferred(Column(Integer), group='dataset')
    result_ag_nays = deferred(Column(Integer), group='dataset')
    result_ag_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_ai_accepted = Column('result_ai_accepted', Integer)
    result_ai_accepted = encoded_property()
    result_ai_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ai_votes_valid = deferred(Column(Integer), group='dataset')
    result_ai_votes_total = deferred(Column(Integer), group='dataset')
    result_ai_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ai_yeas = deferred(Column(Integer), group='dataset')
    result_ai_nays = deferred(Column(Integer), group='dataset')
    result_ai_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_ar_accepted = Column('result_ar_accepted', Integer)
    result_ar_accepted = encoded_property()
    result_ar_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ar_votes_valid = deferred(Column(Integer), group='dataset')
    result_ar_votes_total = deferred(Column(Integer), group='dataset')
    result_ar_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ar_yeas = deferred(Column(Integer), group='dataset')
    result_ar_nays = deferred(Column(Integer), group='dataset')
    result_ar_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_be_accepted = Column('result_be_accepted', Integer)
    result_be_accepted = encoded_property()
    result_be_eligible_voters = deferred(Column(Integer), group='dataset')
    result_be_votes_valid = deferred(Column(Integer), group='dataset')
    result_be_votes_total = deferred(Column(Integer), group='dataset')
    result_be_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_be_yeas = deferred(Column(Integer), group='dataset')
    result_be_nays = deferred(Column(Integer), group='dataset')
    result_be_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_bl_accepted = Column('result_bl_accepted', Integer)
    result_bl_accepted = encoded_property()
    result_bl_eligible_voters = deferred(Column(Integer), group='dataset')
    result_bl_votes_valid = deferred(Column(Integer), group='dataset')
    result_bl_votes_total = deferred(Column(Integer), group='dataset')
    result_bl_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_bl_yeas = deferred(Column(Integer), group='dataset')
    result_bl_nays = deferred(Column(Integer), group='dataset')
    result_bl_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_bs_accepted = Column('result_bs_accepted', Integer)
    result_bs_accepted = encoded_property()
    result_bs_eligible_voters = deferred(Column(Integer), group='dataset')
    result_bs_votes_valid = deferred(Column(Integer), group='dataset')
    result_bs_votes_total = deferred(Column(Integer), group='dataset')
    result_bs_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_bs_yeas = deferred(Column(Integer), group='dataset')
    result_bs_nays = deferred(Column(Integer), group='dataset')
    result_bs_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_fr_accepted = Column('result_fr_accepted', Integer)
    result_fr_accepted = encoded_property()
    result_fr_eligible_voters = deferred(Column(Integer), group='dataset')
    result_fr_votes_valid = deferred(Column(Integer), group='dataset')
    result_fr_votes_total = deferred(Column(Integer), group='dataset')
    result_fr_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_fr_yeas = deferred(Column(Integer), group='dataset')
    result_fr_nays = deferred(Column(Integer), group='dataset')
    result_fr_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_ge_accepted = Column('result_ge_accepted', Integer)
    result_ge_accepted = encoded_property()
    result_ge_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ge_votes_valid = deferred(Column(Integer), group='dataset')
    result_ge_votes_total = deferred(Column(Integer), group='dataset')
    result_ge_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ge_yeas = deferred(Column(Integer), group='dataset')
    result_ge_nays = deferred(Column(Integer), group='dataset')
    result_ge_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_gl_accepted = Column('result_gl_accepted', Integer)
    result_gl_accepted = encoded_property()
    result_gl_eligible_voters = deferred(Column(Integer), group='dataset')
    result_gl_votes_valid = deferred(Column(Integer), group='dataset')
    result_gl_votes_total = deferred(Column(Integer), group='dataset')
    result_gl_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_gl_yeas = deferred(Column(Integer), group='dataset')
    result_gl_nays = deferred(Column(Integer), group='dataset')
    result_gl_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_gr_accepted = Column('result_gr_accepted', Integer)
    result_gr_accepted = encoded_property()
    result_gr_eligible_voters = deferred(Column(Integer), group='dataset')
    result_gr_votes_valid = deferred(Column(Integer), group='dataset')
    result_gr_votes_total = deferred(Column(Integer), group='dataset')
    result_gr_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_gr_yeas = deferred(Column(Integer), group='dataset')
    result_gr_nays = deferred(Column(Integer), group='dataset')
    result_gr_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_ju_accepted = Column('result_ju_accepted', Integer)
    result_ju_accepted = encoded_property()
    result_ju_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ju_votes_valid = deferred(Column(Integer), group='dataset')
    result_ju_votes_total = deferred(Column(Integer), group='dataset')
    result_ju_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ju_yeas = deferred(Column(Integer), group='dataset')
    result_ju_nays = deferred(Column(Integer), group='dataset')
    result_ju_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_lu_accepted = Column('result_lu_accepted', Integer)
    result_lu_accepted = encoded_property()
    result_lu_eligible_voters = deferred(Column(Integer), group='dataset')
    result_lu_votes_valid = deferred(Column(Integer), group='dataset')
    result_lu_votes_total = deferred(Column(Integer), group='dataset')
    result_lu_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_lu_yeas = deferred(Column(Integer), group='dataset')
    result_lu_nays = deferred(Column(Integer), group='dataset')
    result_lu_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_ne_accepted = Column('result_ne_accepted', Integer)
    result_ne_accepted = encoded_property()
    result_ne_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ne_votes_valid = deferred(Column(Integer), group='dataset')
    result_ne_votes_total = deferred(Column(Integer), group='dataset')
    result_ne_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ne_yeas = deferred(Column(Integer), group='dataset')
    result_ne_nays = deferred(Column(Integer), group='dataset')
    result_ne_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_nw_accepted = Column('result_nw_accepted', Integer)
    result_nw_accepted = encoded_property()
    result_nw_eligible_voters = deferred(Column(Integer), group='dataset')
    result_nw_votes_valid = deferred(Column(Integer), group='dataset')
    result_nw_votes_total = deferred(Column(Integer), group='dataset')
    result_nw_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_nw_yeas = deferred(Column(Integer), group='dataset')
    result_nw_nays = deferred(Column(Integer), group='dataset')
    result_nw_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_ow_accepted = Column('result_ow_accepted', Integer)
    result_ow_accepted = encoded_property()
    result_ow_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ow_votes_valid = deferred(Column(Integer), group='dataset')
    result_ow_votes_total = deferred(Column(Integer), group='dataset')
    result_ow_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ow_yeas = deferred(Column(Integer), group='dataset')
    result_ow_nays = deferred(Column(Integer), group='dataset')
    result_ow_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_sg_accepted = Column('result_sg_accepted', Integer)
    result_sg_accepted = encoded_property()
    result_sg_eligible_voters = deferred(Column(Integer), group='dataset')
    result_sg_votes_valid = deferred(Column(Integer), group='dataset')
    result_sg_votes_total = deferred(Column(Integer), group='dataset')
    result_sg_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_sg_yeas = deferred(Column(Integer), group='dataset')
    result_sg_nays = deferred(Column(Integer), group='dataset')
    result_sg_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_sh_accepted = Column('result_sh_accepted', Integer)
    result_sh_accepted = encoded_property()
    result_sh_eligible_voters = deferred(Column(Integer), group='dataset')
    result_sh_votes_valid = deferred(Column(Integer), group='dataset')
    result_sh_votes_total = deferred(Column(Integer), group='dataset')
    result_sh_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_sh_yeas = deferred(Column(Integer), group='dataset')
    result_sh_nays = deferred(Column(Integer), group='dataset')
    result_sh_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_so_accepted = Column('result_so_accepted', Integer)
    result_so_accepted = encoded_property()
    result_so_eligible_voters = deferred(Column(Integer), group='dataset')
    result_so_votes_valid = deferred(Column(Integer), group='dataset')
    result_so_votes_total = deferred(Column(Integer), group='dataset')
    result_so_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_so_yeas = deferred(Column(Integer), group='dataset')
    result_so_nays = deferred(Column(Integer), group='dataset')
    result_so_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_sz_accepted = Column('result_sz_accepted', Integer)
    result_sz_accepted = encoded_property()
    result_sz_eligible_voters = deferred(Column(Integer), group='dataset')
    result_sz_votes_valid = deferred(Column(Integer), group='dataset')
    result_sz_votes_total = deferred(Column(Integer), group='dataset')
    result_sz_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_sz_yeas = deferred(Column(Integer), group='dataset')
    result_sz_nays = deferred(Column(Integer), group='dataset')
    result_sz_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_tg_accepted = Column('result_tg_accepted', Integer)
    result_tg_accepted = encoded_property()
    result_tg_eligible_voters = deferred(Column(Integer), group='dataset')
    result_tg_votes_valid = deferred(Column(Integer), group='dataset')
    result_tg_votes_total = deferred(Column(Integer), group='dataset')
    result_tg_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_tg_yeas = deferred(Column(Integer), group='dataset')
    result_tg_nays = deferred(Column(Integer), group='dataset')
    result_tg_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_ti_accepted = Column('result_ti_accepted', Integer)
    result_ti_accepted = encoded_property()
    result_ti_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ti_votes_valid = deferred(Column(Integer), group='dataset')
    result_ti_votes_total = deferred(Column(Integer), group='dataset')
    result_ti_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ti_yeas = deferred(Column(Integer), group='dataset')
    result_ti_nays = deferred(Column(Integer), group='dataset')
    result_ti_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_ur_accepted = Column('result_ur_accepted', Integer)
    result_ur_accepted = encoded_property()
    result_ur_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ur_votes_valid = deferred(Column(Integer), group='dataset')
    result_ur_votes_total = deferred(Column(Integer), group='dataset')
    result_ur_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ur_yeas = deferred(Column(Integer), group='dataset')
    result_ur_nays = deferred(Column(Integer), group='dataset')
    result_ur_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_vd_accepted = Column('result_vd_accepted', Integer)
    result_vd_accepted = encoded_property()
    result_vd_eligible_voters = deferred(Column(Integer), group='dataset')
    result_vd_votes_valid = deferred(Column(Integer), group='dataset')
    result_vd_votes_total = deferred(Column(Integer), group='dataset')
    result_vd_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_vd_yeas = deferred(Column(Integer), group='dataset')
    result_vd_nays = deferred(Column(Integer), group='dataset')
    result_vd_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_vs_accepted = Column('result_vs_accepted', Integer)
    result_vs_accepted = encoded_property()
    result_vs_eligible_voters = deferred(Column(Integer), group='dataset')
    result_vs_votes_valid = deferred(Column(Integer), group='dataset')
    result_vs_votes_total = deferred(Column(Integer), group='dataset')
    result_vs_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_vs_yeas = deferred(Column(Integer), group='dataset')
    result_vs_nays = deferred(Column(Integer), group='dataset')
    result_vs_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_zg_accepted = Column('result_zg_accepted', Integer)
    result_zg_accepted = encoded_property()
    result_zg_eligible_voters = deferred(Column(Integer), group='dataset')
    result_zg_votes_valid = deferred(Column(Integer), group='dataset')
    result_zg_votes_total = deferred(Column(Integer), group='dataset')
    result_zg_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_zg_yeas = deferred(Column(Integer), group='dataset')
    result_zg_nays = deferred(Column(Integer), group='dataset')
    result_zg_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_zh_accepted = Column('result_zh_accepted', Integer)
    result_zh_accepted = encoded_property()
    result_zh_eligible_voters = deferred(Column(Integer), group='dataset')
    result_zh_votes_valid = deferred(Column(Integer), group='dataset')
    result_zh_votes_total = deferred(Column(Integer), group='dataset')
    result_zh_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_zh_yeas = deferred(Column(Integer), group='dataset')
    result_zh_nays = deferred(Column(Integer), group='dataset')
    result_zh_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    @cached_property
    def results_cantons(self):
        """ Returns the results of all cantons. """

        result = {}
        for canton in Region.cantons():
            value = getattr(self, f'_result_{canton}_accepted')
            if value is not None:
                result.setdefault(value, []).append(Region(canton))

        codes = self.codes('result_accepted')
        return OrderedDict([
            (codes[key], result[key])
            for key in sorted(result.keys())
        ])

    # Authorities
    _department_in_charge = Column('department_in_charge', Integer)
    department_in_charge = encoded_property()
    procedure_number = Column(Numeric(8, 3))
    _position_federal_council = Column('position_federal_council', Integer)
    position_federal_council = encoded_property()
    _position_parliament = Column('position_parliament', Integer)
    position_parliament = encoded_property()
    _position_national_council = Column('position_national_council', Integer)
    position_national_council = encoded_property()
    position_national_council_yeas = Column(Integer)
    position_national_council_nays = Column(Integer)
    _position_council_of_states = Column('position_council_of_states', Integer)
    position_council_of_states = encoded_property()
    position_council_of_states_yeas = Column(Integer)
    position_council_of_states_nays = Column(Integer)

    # Duration
    duration_federal_assembly = Column(Integer)
    duration_post_federal_assembly = Column(Integer)
    duration_initative_collection = Column(Integer)
    duration_initative_federal_council = Column(Integer)
    duration_initative_total = Column(Integer)
    duration_referendum_collection = Column(Integer)
    duration_referendum_total = Column(Integer)
    signatures_valid = Column(Integer)
    signatures_invalid = Column(Integer)

    # Voting recommendations
    recommendations = Column(JSON, nullable=False, default=dict)
    recommendations_other_yes = Column(Text)
    recommendations_other_no = Column(Text)
    recommendations_other_free = Column(Text)
    recommendations_divergent = Column(JSON, nullable=False, default=dict)

    def get_recommendation(self, name):
        """ Get the recommendations by name. """
        return self.codes('recommendation').get(
            self.recommendations.get(name)
        )

    def get_recommendation_of_existing_parties(self):
        """ Get only the existing parties as when this vote was conducted """
        if not self.recommendations:
            return {}
        return {
            k: v for k, v in self.recommendations.items()
            if v != self.ORGANIZATION_NO_LONGER_EXISTS
        }

    def group_recommendations(self, recommendations, ignore_unknown=False):
        """ Group the given recommendations by slogan. """

        codes = self.codes('recommendation')
        recommendation_codes = list(codes.keys())

        def by_recommendation(reco):
            return recommendation_codes.index(reco)

        result = {}
        for actor, recommendation in recommendations:
            if recommendation == self.ORGANIZATION_NO_LONGER_EXISTS:
                continue
            if ignore_unknown and recommendation is None:
                continue

            result.setdefault(recommendation, []).append(actor)

        return OrderedDict([
            (codes[key], result[key])
            for key in sorted(result.keys(), key=by_recommendation)
        ])

    @staticmethod
    def normalize_actor_share_suffix(actor):
        """
        One actor is called sps, but the table name is called
        national_council_share_sp.
        TODO: As soon as the table name is adjusted, this can be removed.
        """
        if actor == 'sps':
            return 'sp'
        return actor

    def get_actors_share(self, actor):
        assert isinstance(actor, str), 'Actor must be a string'
        attr = 'national_council_share_' + \
               self.normalize_actor_share_suffix(actor)
        return getattr(self, attr, 0) or 0

    @cached_property
    def sorted_actors_list(self):
        """
         Returns a list of actors of the current vote sorted by:

        1. codes for recommendations (strength table)
        2. by electoral share (descending)

        It filters out those parties who have no electoral share

        """
        result = []
        for slogan, actor_list in self.recommendations_parties.items():
            actors = (d.name for d in actor_list)
            # Filter out those who have None as share

            result.extend(
                sorted(actors, key=self.get_actors_share, reverse=True)
            )
        return result

    @cached_property
    def recommendations_parties(self):
        """ The recommendations of the parties grouped by slogans. """

        recommendations = self.recommendations or {}
        return self.group_recommendations((
            (Actor(name), recommendations.get(name))
            for name in Actor.parties()
        ))

    @cached_property
    def recommendations_divergent_parties(self, ignore_unknown=True):
        """ The divergent recommendations of the parties grouped by slogans.

        """

        recommendations = self.recommendations_divergent or {}
        return self.group_recommendations((
            (
                (Actor(name.split('_')[0]), Region(name.split('_')[1])),
                recommendation,
            )
            for name, recommendation in sorted(recommendations.items())
        ), ignore_unknown=ignore_unknown)

    @cached_property
    def recommendations_associations(self):
        """ The recommendations of the associations grouped by slogans. """

        def as_list(value, code):
            return [
                (Actor(name.strip()), code)
                for name in (value or '').split(',')
                if name.strip()
            ]

        recommendations = self.recommendations or {}
        recommendations = [
            (Actor(name), recommendations.get(name))
            for name in Actor.associations()
        ]
        recommendations.extend(as_list(self.recommendations_other_yes, 1))
        recommendations.extend(as_list(self.recommendations_other_no, 2))
        recommendations.extend(as_list(self.recommendations_other_free, 5))
        return self.group_recommendations(recommendations)

    # Electoral strength
    national_council_election_year = Column(Integer)
    national_council_share_fdp = Column(Numeric(13, 10))
    national_council_share_cvp = Column(Numeric(13, 10))
    national_council_share_sp = Column(Numeric(13, 10))
    national_council_share_svp = Column(Numeric(13, 10))
    national_council_share_lps = Column(Numeric(13, 10))
    national_council_share_ldu = Column(Numeric(13, 10))
    national_council_share_evp = Column(Numeric(13, 10))
    national_council_share_csp = Column(Numeric(13, 10))
    national_council_share_pda = Column(Numeric(13, 10))
    national_council_share_poch = Column(Numeric(13, 10))
    national_council_share_gps = Column(Numeric(13, 10))
    national_council_share_sd = Column(Numeric(13, 10))
    national_council_share_rep = Column(Numeric(13, 10))
    national_council_share_edu = Column(Numeric(13, 10))
    national_council_share_fps = Column(Numeric(13, 10))
    national_council_share_lega = Column(Numeric(13, 10))
    national_council_share_kvp = Column(Numeric(13, 10))
    national_council_share_glp = Column(Numeric(13, 10))
    national_council_share_bdp = Column(Numeric(13, 10))
    national_council_share_mcg = Column(Numeric(13, 10))
    national_council_share_ubrige = Column(Numeric(13, 10))
    national_council_share_yeas = Column(Numeric(13, 10))
    national_council_share_nays = Column(Numeric(13, 10))
    national_council_share_none = Column(Numeric(13, 10))
    national_council_share_empty = Column(Numeric(13, 10))
    national_council_share_free_vote = Column(Numeric(13, 10))
    national_council_share_neutral = Column(Numeric(13, 10))
    national_council_share_unknown = Column(Numeric(13, 10))

    @cached_property
    def has_national_council_share_data(self):
        if self.national_council_election_year:
            return True
        return False

    # attachments
    voting_text = LocalizedFile()
    brief_description = LocalizedFile()
    federal_council_message = LocalizedFile()
    parliamentary_debate = LocalizedFile()
    voting_booklet = LocalizedFile()
    resolution = LocalizedFile()
    realization = LocalizedFile()
    ad_analysis = LocalizedFile()
    results_by_domain = LocalizedFile()

    # searchable attachment texts
    searchable_text_de_CH = deferred(Column(TSVECTOR))
    searchable_text_fr_CH = deferred(Column(TSVECTOR))

    indexed_files = {
        'voting_text',
        'brief_description',
        'federal_council_message',
        'parliamentary_debate',
        # we don't include the voting booklet, resolution and ad analysis
        # - they might contain other votes from the same day!
        'realization'
    }

    def vectorize_files(self):
        """ Extract the text from the indexed files and store it. """

        for locale, language in (('de_CH', 'german'), ('fr_CH', 'french')):
            files = [
                SwissVote.__dict__[file].__get_by_locale__(self, locale)
                for file in self.indexed_files
            ]
            text = ' '.join([
                extract_pdf_info(file.reference.file)[1] or ''
                for file in files if file
            ]).strip()
            setattr(
                self,
                f'searchable_text_{locale}',
                func.to_tsvector(language, text)
            )

    @observes('files')
    def files_observer(self, files):
        self.vectorize_files()

    def get_file(self, name):
        """ Returns the requested localized file, falls back to the default
        locale.

        """
        fallback = SwissVote.__dict__.get(name).__get_by_locale__(
            self, self.session_manager.default_locale
        )
        return getattr(self, name, None) or fallback



from io import BytesIO
from decimal import Decimal
from datetime import date
from psycopg2.extras import NumericRange
from onegov_testing.utils import create_app
from onegov.core.crypto import hash_password
from onegov.core.crypto import random_token
from onegov.file.utils import as_fileintent
from onegov.pdf import Pdf
from onegov.swissvotes import SwissvotesApp
from onegov.swissvotes.models import Principal, SwissVote
from onegov.swissvotes.models import SwissVoteFile
from onegov.user import User
from pytest import fixture
from transaction import commit
from xlsxwriter.workbook import Workbook


def create_swissvotes_app(request, temporary_path):
    app = create_app(
        SwissvotesApp,
        request,
        use_smtp=True,
        depot_backend='depot.io.local.LocalFileStorage',
        depot_storage_path=str(temporary_path),
    )
    app.session_manager.set_locale('de_CH', 'de_CH')

    session = app.session()
    session.add(User(
        username='admin@example.org',
        password_hash=request.getfixturevalue('swissvotes_password'),
        role='admin'
    ))
    session.add(User(
        realname='Publisher',
        username='publisher@example.org',
        password_hash=request.getfixturevalue('swissvotes_password'),
        role='editor'
    ))
    session.add(User(
        realname='Editor',
        username='editor@example.org',
        password_hash=request.getfixturevalue('swissvotes_password'),
        role='member'
    ))

    commit()
    return app


@fixture(scope='session')
def swissvotes_password():
    return hash_password('hunter2')


@fixture(scope="function")
def principal():
    yield Principal()


@fixture(scope="function")
def swissvotes_app(request, temporary_path):
    app = create_swissvotes_app(request, temporary_path)
    yield app
    app.session_manager.dispose()


@fixture(scope="function")
def attachments(swissvotes_app):
    result = {}
    for name, content in (
        ('ad_analysis', "Inserateanalyse"),
        ('brief_description', "Kurschbeschreibung"),
        ('federal_council_message', "Message du Conseil fédéral"),
        ('parliamentary_debate', "Parlamentdebatte"),
        ('realization', "Réalisation"),
        ('resolution', "Arrêté constatant le résultat"),
        ('voting_booklet', "Brochure explicative"),
        ('voting_text', "Abstimmungstext"),
    ):
        file = BytesIO()
        pdf = Pdf(file)
        pdf.init_report()
        pdf.p(content)
        pdf.generate()
        file.seek(0)

        attachment = SwissVoteFile(id=random_token())
        attachment.reference = as_fileintent(file, name)
        result[name] = attachment

    for name in (
        'results_by_domain',
    ):
        file = BytesIO()
        workbook = Workbook(file)
        worksheet = workbook.add_worksheet('DATA')
        worksheet.write_row(0, 0, ['a', 'b'])
        worksheet.write_row(1, 0, [100, 200])
        workbook.close()
        file.seek(0)

        attachment = SwissVoteFile(id=random_token())
        attachment.reference = as_fileintent(file, name)
        result[name] = attachment

    yield result


@fixture(scope="function")
def postgres_version(session):
    connection = session.connection()
    yield connection.execute('show server_version;').fetchone()[0]


@fixture(scope="function")
def sample_vote():
    vote = SwissVote()
    vote.bfs_number = Decimal('100.1')
    vote.date = date(1990, 6, 2)
    vote.legislation_number = 4
    vote.legislation_decade = NumericRange(1990, 1994)
    vote.title_de = "Vote DE"
    vote.title_fr = "Vote FR"
    vote.short_title_de = "V D"
    vote.short_title_fr = "V F"
    vote.keyword = "Keyword"
    vote.votes_on_same_day = 2
    vote._legal_form = 1
    vote.initiator = "Initiator"
    vote.anneepolitique = "anneepolitique"
    vote.bfs_map_de = (
        "https://www.atlas.bfs.admin.ch/maps/12/map/mapIdOnly/1815_de.html"
    )
    vote.bfs_map_fr = "htt(ps://www.ap/mapIdOnly/1815[e.html}"
    vote.descriptor_1_level_1 = Decimal('4')
    vote.descriptor_1_level_2 = Decimal('4.2')
    vote.descriptor_1_level_3 = Decimal('4.21')
    vote.descriptor_2_level_1 = Decimal('10')
    vote.descriptor_2_level_2 = Decimal('10.3')
    vote.descriptor_2_level_3 = Decimal('10.35')
    vote.descriptor_3_level_1 = Decimal('10')
    vote.descriptor_3_level_2 = Decimal('10.3')
    vote.descriptor_3_level_3 = Decimal('10.33')
    vote._result = 1
    vote.result_eligible_voters = 2
    vote.result_votes_empty = 3
    vote.result_votes_invalid = 4
    vote.result_votes_valid = 5
    vote.result_votes_total = 6
    vote.result_turnout = Decimal('20.01')
    vote._result_people_accepted = 1
    vote.result_people_yeas = 8
    vote.result_people_nays = 9
    vote.result_people_yeas_p = Decimal('40.01')
    vote._result_cantons_accepted = 1
    vote.result_cantons_yeas = Decimal('1.5')
    vote.result_cantons_nays = Decimal('24.5')
    vote.result_cantons_yeas_p = Decimal('60.01')
    vote.result_ag_eligible_voters = 101
    vote.result_ag_votes_valid = 102
    vote.result_ag_votes_total = 103
    vote.result_ag_turnout = Decimal('10.40')
    vote.result_ag_yeas = 105
    vote.result_ag_nays = 107
    vote.result_ag_yeas_p = Decimal('10.80')
    vote._result_ag_accepted = 0
    vote.result_ai_eligible_voters = 101
    vote.result_ai_votes_valid = 102
    vote.result_ai_votes_total = 103
    vote.result_ai_turnout = Decimal('10.40')
    vote.result_ai_yeas = 105
    vote.result_ai_nays = 107
    vote.result_ai_yeas_p = Decimal('10.80')
    vote._result_ai_accepted = 0
    vote.result_ar_eligible_voters = 101
    vote.result_ar_votes_valid = 102
    vote.result_ar_votes_total = 103
    vote.result_ar_turnout = Decimal('10.40')
    vote.result_ar_yeas = 105
    vote.result_ar_nays = 107
    vote.result_ar_yeas_p = Decimal('10.80')
    vote._result_ar_accepted = 0
    vote.result_be_eligible_voters = 101
    vote.result_be_votes_valid = 102
    vote.result_be_votes_total = 103
    vote.result_be_turnout = Decimal('10.40')
    vote.result_be_yeas = 105
    vote.result_be_nays = 107
    vote.result_be_yeas_p = Decimal('10.80')
    vote._result_be_accepted = 0
    vote.result_bl_eligible_voters = 101
    vote.result_bl_votes_valid = 102
    vote.result_bl_votes_total = 103
    vote.result_bl_turnout = Decimal('10.40')
    vote.result_bl_yeas = 105
    vote.result_bl_nays = 107
    vote.result_bl_yeas_p = Decimal('10.80')
    vote._result_bl_accepted = 0
    vote.result_bs_eligible_voters = 101
    vote.result_bs_votes_valid = 102
    vote.result_bs_votes_total = 103
    vote.result_bs_turnout = Decimal('10.40')
    vote.result_bs_yeas = 105
    vote.result_bs_nays = 107
    vote.result_bs_yeas_p = Decimal('10.80')
    vote._result_bs_accepted = 0
    vote.result_fr_eligible_voters = 101
    vote.result_fr_votes_valid = 102
    vote.result_fr_votes_total = 103
    vote.result_fr_turnout = Decimal('10.40')
    vote.result_fr_yeas = 105
    vote.result_fr_nays = 107
    vote.result_fr_yeas_p = Decimal('10.80')
    vote._result_fr_accepted = 0
    vote.result_ge_eligible_voters = 101
    vote.result_ge_votes_valid = 102
    vote.result_ge_votes_total = 103
    vote.result_ge_turnout = Decimal('10.40')
    vote.result_ge_yeas = 105
    vote.result_ge_nays = 107
    vote.result_ge_yeas_p = Decimal('10.80')
    vote._result_ge_accepted = 0
    vote.result_gl_eligible_voters = 101
    vote.result_gl_votes_valid = 102
    vote.result_gl_votes_total = 103
    vote.result_gl_turnout = Decimal('10.40')
    vote.result_gl_yeas = 105
    vote.result_gl_nays = 107
    vote.result_gl_yeas_p = Decimal('10.80')
    vote._result_gl_accepted = 0
    vote.result_gr_eligible_voters = 101
    vote.result_gr_votes_valid = 102
    vote.result_gr_votes_total = 103
    vote.result_gr_turnout = Decimal('10.40')
    vote.result_gr_yeas = 105
    vote.result_gr_nays = 107
    vote.result_gr_yeas_p = Decimal('10.80')
    vote._result_gr_accepted = 0
    vote.result_ju_eligible_voters = 101
    vote.result_ju_votes_valid = 102
    vote.result_ju_votes_total = 103
    vote.result_ju_turnout = Decimal('10.40')
    vote.result_ju_yeas = 105
    vote.result_ju_nays = 107
    vote.result_ju_yeas_p = Decimal('10.80')
    vote._result_ju_accepted = 0
    vote.result_lu_eligible_voters = 101
    vote.result_lu_votes_valid = 102
    vote.result_lu_votes_total = 103
    vote.result_lu_turnout = Decimal('10.40')
    vote.result_lu_yeas = 105
    vote.result_lu_nays = 107
    vote.result_lu_yeas_p = Decimal('10.80')
    vote._result_lu_accepted = 0
    vote.result_ne_eligible_voters = 101
    vote.result_ne_votes_valid = 102
    vote.result_ne_votes_total = 103
    vote.result_ne_turnout = Decimal('10.40')
    vote.result_ne_yeas = 105
    vote.result_ne_nays = 107
    vote.result_ne_yeas_p = Decimal('10.80')
    vote._result_ne_accepted = 0
    vote.result_nw_eligible_voters = 101
    vote.result_nw_votes_valid = 102
    vote.result_nw_votes_total = 103
    vote.result_nw_turnout = Decimal('10.40')
    vote.result_nw_yeas = 105
    vote.result_nw_nays = 107
    vote.result_nw_yeas_p = Decimal('10.80')
    vote._result_nw_accepted = 0
    vote.result_ow_eligible_voters = 101
    vote.result_ow_votes_valid = 102
    vote.result_ow_votes_total = 103
    vote.result_ow_turnout = Decimal('10.40')
    vote.result_ow_yeas = 105
    vote.result_ow_nays = 107
    vote.result_ow_yeas_p = Decimal('10.80')
    vote._result_ow_accepted = 0
    vote.result_sg_eligible_voters = 101
    vote.result_sg_votes_valid = 102
    vote.result_sg_votes_total = 103
    vote.result_sg_turnout = Decimal('10.40')
    vote.result_sg_yeas = 105
    vote.result_sg_nays = 107
    vote.result_sg_yeas_p = Decimal('10.80')
    vote._result_sg_accepted = 0
    vote.result_sh_eligible_voters = 101
    vote.result_sh_votes_valid = 102
    vote.result_sh_votes_total = 103
    vote.result_sh_turnout = Decimal('10.40')
    vote.result_sh_yeas = 105
    vote.result_sh_nays = 107
    vote.result_sh_yeas_p = Decimal('10.80')
    vote._result_sh_accepted = 0
    vote.result_so_eligible_voters = 101
    vote.result_so_votes_valid = 102
    vote.result_so_votes_total = 103
    vote.result_so_turnout = Decimal('10.40')
    vote.result_so_yeas = 105
    vote.result_so_nays = 107
    vote.result_so_yeas_p = Decimal('10.80')
    vote._result_so_accepted = 0
    vote.result_sz_eligible_voters = 101
    vote.result_sz_votes_valid = 102
    vote.result_sz_votes_total = 103
    vote.result_sz_turnout = Decimal('10.40')
    vote.result_sz_yeas = 105
    vote.result_sz_nays = 107
    vote.result_sz_yeas_p = Decimal('10.80')
    vote._result_sz_accepted = 0
    vote.result_tg_eligible_voters = 101
    vote.result_tg_votes_valid = 102
    vote.result_tg_votes_total = 103
    vote.result_tg_turnout = Decimal('10.40')
    vote.result_tg_yeas = 105
    vote.result_tg_nays = 107
    vote.result_tg_yeas_p = Decimal('10.80')
    vote._result_tg_accepted = 0
    vote.result_ti_eligible_voters = 101
    vote.result_ti_votes_valid = 102
    vote.result_ti_votes_total = 103
    vote.result_ti_turnout = Decimal('10.40')
    vote.result_ti_yeas = 105
    vote.result_ti_nays = 107
    vote.result_ti_yeas_p = Decimal('10.80')
    vote._result_ti_accepted = 0
    vote.result_ur_eligible_voters = 101
    vote.result_ur_votes_valid = 102
    vote.result_ur_votes_total = 103
    vote.result_ur_turnout = Decimal('10.40')
    vote.result_ur_yeas = 105
    vote.result_ur_nays = 107
    vote.result_ur_yeas_p = Decimal('10.80')
    vote._result_ur_accepted = 0
    vote.result_vd_eligible_voters = 101
    vote.result_vd_votes_valid = 102
    vote.result_vd_votes_total = 103
    vote.result_vd_turnout = Decimal('10.40')
    vote.result_vd_yeas = 105
    vote.result_vd_nays = 107
    vote.result_vd_yeas_p = Decimal('10.80')
    vote._result_vd_accepted = 1
    vote.result_vs_eligible_voters = 101
    vote.result_vs_votes_valid = 102
    vote.result_vs_votes_total = 103
    vote.result_vs_turnout = Decimal('10.40')
    vote.result_vs_yeas = 105
    vote.result_vs_nays = 107
    vote.result_vs_yeas_p = Decimal('10.80')
    vote._result_vs_accepted = 1
    vote.result_zg_eligible_voters = 101
    vote.result_zg_votes_valid = 102
    vote.result_zg_votes_total = 103
    vote.result_zg_turnout = Decimal('10.40')
    vote.result_zg_yeas = 105
    vote.result_zg_nays = 107
    vote.result_zg_yeas_p = Decimal('10.80')
    vote._result_zg_accepted = 0
    vote.result_zh_eligible_voters = 101
    vote.result_zh_votes_valid = 102
    vote.result_zh_votes_total = 103
    vote.result_zh_turnout = Decimal('10.40')
    vote.result_zh_yeas = 105
    vote.result_zh_nays = 107
    vote.result_zh_yeas_p = Decimal('10.80')
    vote._department_in_charge = 1
    vote.procedure_number = Decimal('24.557')
    vote._position_federal_council = 1
    vote._position_parliament = 1
    vote._position_national_council = 1
    vote.position_national_council_yeas = 10
    vote.position_national_council_nays = 20
    vote._position_council_of_states = 1
    vote.position_council_of_states_yeas = 30
    vote.position_council_of_states_nays = 40
    vote.duration_federal_assembly = 30
    vote.duration_post_federal_assembly = 31
    vote.duration_initative_collection = 32
    vote.duration_initative_federal_council = 33
    vote.duration_initative_total = 34
    vote.duration_referendum_collection = 35
    vote.duration_referendum_total = 36
    vote.signatures_valid = 40
    vote.signatures_invalid = 41
    vote.recommendations = {
        'fdp': 1,
        'cvp': 1,
        'sps': 1,
        'svp': 1,
        'lps': 2,
        'ldu': 2,
        'evp': 2,
        'csp': 3,
        'pda': 3,
        'poch': 3,
        'gps': 4,
        'sd': 4,
        'rep': 4,
        'edu': 5,
        'fps': 5,
        'lega': 5,
        'kvp': 66,
        'glp': 66,
        'bdp': None,
        'mcg': 9999,
        'sav': 1,
        'eco': 2,
        'sgv': 3,
        'sbv-usp': 3,
        'sgb': 3,
        'travs': 3,
        'vsa': 9999,
        'vpod': 1,
        'ssv': 1,
        'gem': 1,
        'kdk': 1,
        'vdk': 1,
        'endk': 1,
        'fdk': 1,
        'edk': 1,
        'gdk': 1,
        'ldk': 1,
        'sodk': 1,
        'kkjpd': 1,
        'bpuk': 1,
        'sbk': 1,
        'acs': 1,
        'tcs': 1,
        'vcs': 1,
        'voev': 1
    }
    vote.recommendations_other_yes = "Pro Velo"
    vote.recommendations_other_no = None
    vote.recommendations_other_free = "Pro Natura, Greenpeace"
    vote.recommendations_divergent = {
        'edu_vso': 1,
        'fdp_ti': 1,
        'fdp-fr_ch': 2,
        'jcvp_ch': 2,
    }
    vote.national_council_election_year = 1990
    vote.national_council_share_fdp = Decimal('01.10')
    vote.national_council_share_cvp = Decimal('02.10')
    vote.national_council_share_sp = Decimal('03.10')
    vote.national_council_share_svp = Decimal('04.10')
    vote.national_council_share_lps = Decimal('05.10')
    vote.national_council_share_ldu = Decimal('06.10')
    vote.national_council_share_evp = Decimal('07.10')
    vote.national_council_share_csp = Decimal('08.10')
    vote.national_council_share_pda = Decimal('09.10')
    vote.national_council_share_poch = Decimal('10.10')
    vote.national_council_share_gps = Decimal('11.10')
    vote.national_council_share_sd = Decimal('12.10')
    vote.national_council_share_rep = Decimal('13.10')
    vote.national_council_share_edu = Decimal('14.10')
    vote.national_council_share_fps = Decimal('15.10')
    vote.national_council_share_lega = Decimal('16.10')
    vote.national_council_share_kvp = Decimal('17.10')
    vote.national_council_share_glp = Decimal('18.10')
    vote.national_council_share_bdp = Decimal('19.10')
    vote.national_council_share_mcg = Decimal('20.20')
    vote.national_council_share_ubrige = Decimal('21.20')
    vote.national_council_share_yeas = Decimal('22.20')
    vote.national_council_share_nays = Decimal('23.20')
    vote.national_council_share_neutral = Decimal('24.20')
    vote.national_council_share_none = Decimal('25.20')
    vote.national_council_share_empty = Decimal('26.20')
    vote.national_council_share_free_vote = Decimal('27.20')
    vote.national_council_share_unknown = Decimal('28.20')
    vote.national_council_share_vague = Decimal('28.20')
    return vote



from onegov.form import Form
from onegov.form.fields import MultiCheckboxField
from onegov.swissvotes import _
from onegov.swissvotes.collections import SwissVoteCollection
from onegov.swissvotes.fields import PolicyAreaField
from onegov.swissvotes.models import PolicyArea
from onegov.swissvotes.models import PolicyAreaDefinition
from onegov.swissvotes.models.vote import SwissVote
from wtforms import HiddenField
from wtforms import RadioField
from wtforms import StringField
from wtforms.fields.html5 import DateField


class SearchForm(Form):

    from_date = DateField(
        label=_("From date"),
        render_kw={'size': 6, 'clear': False}
    )

    to_date = DateField(
        label=_("To date"),
        render_kw={'size': 6}
    )

    legal_form = MultiCheckboxField(
        label=_("Legal form"),
        coerce=int,
        render_kw={'size': 6, 'clear': False}
    )

    result = MultiCheckboxField(
        label=_("Voting result"),
        coerce=int,
        render_kw={'size': 6}
    )

    policy_area = PolicyAreaField(
        label=_("Policy area"),
        choices=[],
        render_kw={'size': 6, 'clear': False}
    )

    term = StringField(
        label=_("Text Search"),
        render_kw={'size': 4, 'clear': False},
    )

    full_text = RadioField(
        label=_("Full Text"),
        choices=(
            (1, _("Yes")),
            (0, _("No")),
        ),
        coerce=bool,
        render_kw={'size': 2},
        description=_(
            "Searches all text fields and attachments. Select «No» to limit "
            "the search to title, keyword, vote number, and procedure number."
        )
    )

    position_federal_council = MultiCheckboxField(
        label=_("Position of the Federal Council"),
        fieldset=_("Other Filters"),
        coerce=int,
        render_kw={'size': 3, 'clear': False}
    )

    position_national_council = MultiCheckboxField(
        label=_("Position of the National Council"),
        fieldset=_("Other Filters"),
        coerce=int,
        render_kw={'size': 3, 'clear': False}
    )

    position_council_of_states = MultiCheckboxField(
        label=_("Position of the Council of States"),
        fieldset=_("Other Filters"),
        coerce=int,
        render_kw={'size': 3}
    )

    sort_by = HiddenField()
    sort_order = HiddenField()

    def populate_policy_area(self):
        votes = SwissVoteCollection(self.request.app)
        available = votes.available_descriptors

        def serialize(item):
            children = [serialize(child) for child in item.children]
            children = [child for child in children if child]
            if not item.value:
                return children
            value = PolicyArea(item.path).descriptor_decimal
            if not children and not any((value in s for s in available)):
                return None
            return {
                "label": self.request.translate(item.label),
                "value": '.'.join([str(x) for x in item.path]),
                "children": children
            }

        self.policy_area.tree = serialize(PolicyAreaDefinition.all())

    def populate_choice(self, name, add_none=False):
        field = getattr(self, name)
        field.choices = list(SwissVote.codes(name).items())
        if add_none:
            field.choices.append((-1, _("Without / Unknown")))

    def on_request(self):
        if hasattr(self, 'csrf_token'):
            self.delete_field('csrf_token')
        self.populate_choice('legal_form')
        self.populate_choice('result')
        self.populate_choice('position_federal_council')
        self.populate_choice('position_national_council')
        self.populate_choice('position_council_of_states')
        self.populate_policy_area()

    def select_all(self, name):
        field = getattr(self, name)
        if not field.data:
            field.data = list(next(zip(*field.choices)))

    def apply_model(self, model):
        self.from_date.data = model.from_date
        self.to_date.data = model.to_date
        self.legal_form.data = model.legal_form
        self.result.data = model.result
        self.policy_area.data = model.policy_area
        self.term.data = model.term
        self.full_text.data = 0 if model.full_text is False else 1
        self.position_federal_council.data = model.position_federal_council
        self.position_national_council.data = model.position_national_council
        self.position_council_of_states.data = model.position_council_of_states
        self.sort_by.data = model.sort_by
        self.sort_order.data = model.sort_order

        # default unselected checkboxes to all choices
        self.select_all('legal_form')
        self.select_all('result')
        self.select_all('position_federal_council')
        self.select_all('position_national_council')
        self.select_all('position_council_of_states')



from cached_property import cached_property
from decimal import Decimal
from onegov.swissvotes import _


class PolicyArea(object):
    """ Helper class for handling of descriptors.

    There are three levels of descriptors, each new level refining the
    given category.

    Policy areas are internally represented as a string value containing the
    descriptor of all three levels separated by a comma, e.g. "1.12.121".

    Policy areas are stored in the dataset as float, with the pre-decimal part
    refering to the first level category and the decimal part to the category
    of the given level. For example:

        Level 1 descriptor "1": 1
        Level 2 descriptor "1.12": 1.12
        Level 3 descriptor "1.12.121": 1.121

    """

    def __init__(self, value, level=None):
        """ Creates a new policy descriptor out of the given value.

        The given value might be a string (such as "1.12" or "1.12.121"), a
        list (such as [1, 12] or [1, 12, 121]) or a float together with a level
        (such as 1.12/2 or 1.121/3).

        """

        if isinstance(value, str):
            self.value = value
        elif isinstance(value, list):
            self.value = '.'.join([str(x) for x in value])
        elif isinstance(value, Decimal):
            assert level is not None
            self.value = '.'.join(
                str(int(value * 10 ** x)) for x in range(level)
            )
        else:
            raise NotImplementedError()

    def __repr__(self):
        return self.value

    def __eq__(self, other):
        return self.value == other.value

    @cached_property
    def level(self):
        return self.value.count('.') + 1

    @cached_property
    def descriptor(self):
        """ Returns the highest descriptor, e.g. 121 if "1.12.121". """

        return int(self.value.split('.')[-1])

    @cached_property
    def descriptor_path(self):
        """ Returns all descriptors, e.g [1, 12, 12, 121] if "1.12.121". """

        return [int(part) for part in self.value.split('.')]

    @cached_property
    def descriptor_decimal(self):
        """ Returns the descriptor as float for the dataset, e.g 1.121 if
        "1.12.121".

        """

        return Decimal(self.descriptor) / (10 ** (self.level - 1))

    @cached_property
    def label(self):
        """ Returns a translatable label of the highest descriptor, e.g.
        "Bundesverfassung" if "1.12.121".

        """

        return self.label_path[-1]

    @cached_property
    def label_path(self):
        """ Returns translatable labels for all descriptor levels, e.g.
        ["Staatsordnung", "Politisches System", "Bundesverfassung"] if
        "1.12.121".

        """

        result = []
        lookup = PolicyAreaDefinition.all()
        for part in self.descriptor_path:
            lookup = lookup.get(part)
            if not lookup:
                result.append(str(self.descriptor_decimal))
                break
            result.append(lookup.label or str(self.descriptor))
        return result

    def html(self, request):
        title = ' &gt; '.join([
            request.translate(part) for part in self.label_path
        ])
        return f'<span>{title}</span>'


class PolicyAreaDefinition(object):
    """ Helper class for all the policy areas and their translatable labels.

    Example: Get the label of the policy area "1.12.121":

        PolicyAreaDefinition.all().get(1).get(12).get(121).label

    """

    def __init__(self, path=None, label=None, children=None):
        self.path = path or []
        self.decimal = None
        self.value = path[-1] if self.path else None
        self.label = label
        self.children = children or []
        self.index = {
            child.value: index for index, child in enumerate(self.children)
        }

    def get(self, key):
        """ Returns the child with the given value. """
        if key in self.index:
            return self.children[self.index[key]]

    @staticmethod
    def all():
        """ Returns the tree of all policy areas. """

        return PolicyAreaDefinition(children=[
            PolicyAreaDefinition([1], _("d-1-1"), [
                PolicyAreaDefinition([1, 11], _("d-2-11")),
                PolicyAreaDefinition([1, 12], _("d-2-12"), [
                    PolicyAreaDefinition([1, 12, 121], _("d-3-121")),
                    PolicyAreaDefinition([1, 12, 122], _("d-3-122")),
                    PolicyAreaDefinition([1, 12, 123], _("d-3-123")),
                    PolicyAreaDefinition([1, 12, 124], _("d-3-124")),
                ]),
                PolicyAreaDefinition([1, 13], _("d-2-13"), [
                    PolicyAreaDefinition([1, 13, 131], _("d-3-131")),
                    PolicyAreaDefinition([1, 13, 132], _("d-3-132")),
                    PolicyAreaDefinition([1, 13, 133], _("d-3-133")),
                    PolicyAreaDefinition([1, 13, 134], _("d-3-134")),
                ]),
                PolicyAreaDefinition([1, 14], _("d-2-14"), [
                    PolicyAreaDefinition([1, 14, 141], _("d-3-141")),
                    PolicyAreaDefinition([1, 14, 142], _("d-3-142")),
                    PolicyAreaDefinition([1, 14, 143], _("d-3-143")),
                ]),
                PolicyAreaDefinition([1, 15], _("d-2-15"), [
                    PolicyAreaDefinition([1, 15, 151], _("d-3-151")),
                    PolicyAreaDefinition([1, 15, 152], _("d-3-152")),
                    PolicyAreaDefinition([1, 15, 153], _("d-3-153")),
                ]),
                PolicyAreaDefinition([1, 16], _("d-2-16"), [
                    PolicyAreaDefinition([1, 16, 161], _("d-3-161")),
                    PolicyAreaDefinition([1, 16, 162], _("d-3-162")),
                    PolicyAreaDefinition([1, 16, 163], _("d-3-163")),
                    PolicyAreaDefinition([1, 16, 164], _("d-3-164")),
                    PolicyAreaDefinition([1, 16, 165], _("d-3-165")),
                    PolicyAreaDefinition([1, 16, 166], _("d-3-166")),
                ])
            ]),
            PolicyAreaDefinition([2], _("d-1-2"), [
                PolicyAreaDefinition([2, 21], _("d-2-21"), [
                    PolicyAreaDefinition([2, 21, 211], _("d-3-211")),
                    PolicyAreaDefinition([2, 21, 212], _("d-3-212")),
                    PolicyAreaDefinition([2, 21, 213], _("d-3-213")),
                ]),
                PolicyAreaDefinition([2, 22], _("d-2-22"), [
                    PolicyAreaDefinition([2, 22, 221], _("d-3-221")),
                    PolicyAreaDefinition([2, 22, 222], _("d-3-222")),
                    PolicyAreaDefinition([2, 22, 223], _("d-3-223")),
                    PolicyAreaDefinition([2, 22, 224], _("d-3-224")),
                ]),
                PolicyAreaDefinition([2, 23], _("d-2-23"), [
                    PolicyAreaDefinition([2, 23, 231], _("d-3-231")),
                    PolicyAreaDefinition([2, 23, 232], _("d-3-232")),
                ]),
                PolicyAreaDefinition([2, 24], _("d-2-24")),
                PolicyAreaDefinition([2, 25], _("d-2-25")),
                PolicyAreaDefinition([2, 26], _("d-2-26"), [
                    PolicyAreaDefinition([2, 26, 261], _("d-3-261")),
                    PolicyAreaDefinition([2, 26, 262], _("d-3-262")),
                ]),
                PolicyAreaDefinition([2, 27], _("d-2-27")),
                PolicyAreaDefinition([2, 28], _("d-2-28")),
            ]),
            PolicyAreaDefinition([3], _("d-1-3"), [
                PolicyAreaDefinition([3, 31], _("d-2-31"), [
                    PolicyAreaDefinition([3, 31, 311], _("d-3-311")),
                    PolicyAreaDefinition([3, 31, 312], _("d-3-312")),
                    PolicyAreaDefinition([3, 31, 313], _("d-3-313")),
                ]),
                PolicyAreaDefinition([3, 32], _("d-2-32"), [
                    PolicyAreaDefinition([3, 32, 321], _("d-3-321")),
                    PolicyAreaDefinition([3, 32, 322], _("d-3-322")),
                    PolicyAreaDefinition([3, 32, 323], _("d-3-323")),
                    PolicyAreaDefinition([3, 32, 324], _("d-3-324")),
                    PolicyAreaDefinition([3, 32, 325], _("d-3-325")),
                    PolicyAreaDefinition([3, 32, 326], _("d-3-326")),
                    PolicyAreaDefinition([3, 32, 327], _("d-3-327")),
                    PolicyAreaDefinition([3, 32, 328], _("d-3-328")),
                ]),
                PolicyAreaDefinition([3, 33], _("d-2-33")),
            ]),
            PolicyAreaDefinition([4], _("d-1-4"), [
                PolicyAreaDefinition([4, 41], _("d-2-41"), [
                    PolicyAreaDefinition([4, 41, 411], _("d-3-411")),
                    PolicyAreaDefinition([4, 41, 412], _("d-3-412")),
                    PolicyAreaDefinition([4, 41, 413], _("d-3-413")),
                    PolicyAreaDefinition([4, 41, 414], _("d-3-414")),
                    PolicyAreaDefinition([4, 41, 415], _("d-3-415")),
                    PolicyAreaDefinition([4, 41, 416], _("d-3-416")),
                ]),
                PolicyAreaDefinition([4, 42], _("d-2-42"), [
                    PolicyAreaDefinition([4, 42, 421], _("d-3-421")),
                    PolicyAreaDefinition([4, 42, 422], _("d-3-422")),
                    PolicyAreaDefinition([4, 42, 423], _("d-3-423")),
                    PolicyAreaDefinition([4, 42, 424], _("d-3-424")),
                ]),
                PolicyAreaDefinition([4, 43], _("d-2-43"), [
                    PolicyAreaDefinition([4, 43, 431], _("d-3-431")),
                    PolicyAreaDefinition([4, 43, 432], _("d-3-432")),
                ]),
                PolicyAreaDefinition([4, 44], _("d-2-44"), [
                    PolicyAreaDefinition([4, 44, 441], _("d-3-441")),
                    PolicyAreaDefinition([4, 44, 442], _("d-3-442")),
                    PolicyAreaDefinition([4, 44, 443], _("d-3-443")),
                ]),
            ]),
            PolicyAreaDefinition([5], _("d-1-5"), [
                PolicyAreaDefinition([5, 51], _("d-2-51")),
                PolicyAreaDefinition([5, 52], _("d-2-52")),
                PolicyAreaDefinition([5, 53], _("d-2-53")),
                PolicyAreaDefinition([5, 54], _("d-2-54")),
                PolicyAreaDefinition([5, 55], _("d-2-55")),
            ]),
            PolicyAreaDefinition([6], _("d-1-6"), [
                PolicyAreaDefinition([6, 61], _("d-2-61"), [
                    PolicyAreaDefinition([6, 61, 611], _("d-3-611")),
                    PolicyAreaDefinition([6, 61, 612], _("d-3-612")),
                    PolicyAreaDefinition([6, 61, 613], _("d-3-613")),
                    PolicyAreaDefinition([6, 61, 614], _("d-3-614")),
                ]),
                PolicyAreaDefinition([6, 62], _("d-2-62")),
                PolicyAreaDefinition([6, 63], _("d-2-63")),
                PolicyAreaDefinition([6, 64], _("d-2-64")),
            ]),
            PolicyAreaDefinition([7], _("d-1-7"), [
                PolicyAreaDefinition([7, 71], _("d-2-71")),
                PolicyAreaDefinition([7, 72], _("d-2-72")),
                PolicyAreaDefinition([7, 73], _("d-2-73")),
                PolicyAreaDefinition([7, 74], _("d-2-74")),
                PolicyAreaDefinition([7, 75], _("d-2-75")),
            ]),
            PolicyAreaDefinition([8], _("d-1-8"), [
                PolicyAreaDefinition([8, 81], _("d-2-81"), [
                    PolicyAreaDefinition([8, 81, 811], _("d-3-811")),
                    PolicyAreaDefinition([8, 81, 812], _("d-3-812")),
                ]),
                PolicyAreaDefinition([8, 82], _("d-2-82"), [
                    PolicyAreaDefinition([8, 82, 821], _("d-3-821")),
                    PolicyAreaDefinition([8, 82, 822], _("d-3-822")),
                ]),
                PolicyAreaDefinition([8, 83], _("d-2-83"), [
                    PolicyAreaDefinition([8, 83, 831], _("d-3-831")),
                    PolicyAreaDefinition([8, 83, 832], _("d-3-832")),
                ]),
                PolicyAreaDefinition([8, 84], _("d-2-84")),
                PolicyAreaDefinition([8, 85], _("d-2-85")),
                PolicyAreaDefinition([8, 86], _("d-2-86")),
                PolicyAreaDefinition([8, 87], _("d-2-87")),
            ]),
            PolicyAreaDefinition([9], _("d-1-9"), [
                PolicyAreaDefinition([9, 91], _("d-2-91"), [
                    PolicyAreaDefinition([9, 91, 911], _("d-3-911")),
                    PolicyAreaDefinition([9, 91, 912], _("d-3-912")),
                ]),
                PolicyAreaDefinition([9, 92], _("d-2-92"), [
                    PolicyAreaDefinition([9, 92, 921], _("d-3-921")),
                    PolicyAreaDefinition([9, 92, 922], _("d-3-922")),
                ]),
                PolicyAreaDefinition([9, 93], _("d-2-93"), [
                    PolicyAreaDefinition([9, 93, 931], _("d-3-931")),
                    PolicyAreaDefinition([9, 93, 932], _("d-3-932")),
                    PolicyAreaDefinition([9, 93, 933], _("d-3-933")),
                    PolicyAreaDefinition([9, 93, 934], _("d-3-934")),
                    PolicyAreaDefinition([9, 93, 935], _("d-3-935")),
                    PolicyAreaDefinition([9, 93, 936], _("d-3-936")),
                    PolicyAreaDefinition([9, 93, 937], _("d-3-937")),
                    PolicyAreaDefinition([9, 93, 938], _("d-3-938")),
                ]),
            ]),
            PolicyAreaDefinition([10], _("d-1-10"), [
                PolicyAreaDefinition([10, 101], _("d-2-101"), [
                    PolicyAreaDefinition([10, 101, 1011], _("d-3-1011")),
                    PolicyAreaDefinition([10, 101, 1012], _("d-3-1012")),
                    PolicyAreaDefinition([10, 101, 1013], _("d-3-1013")),
                    PolicyAreaDefinition([10, 101, 1014], _("d-3-1014")),
                    PolicyAreaDefinition([10, 101, 1015], _("d-3-1015")),
                ]),
                PolicyAreaDefinition([10, 102], _("d-2-102"), [
                    PolicyAreaDefinition([10, 102, 1021], _("d-3-1021")),
                    PolicyAreaDefinition([10, 102, 1022], _("d-3-1022")),
                    PolicyAreaDefinition([10, 102, 1023], _("d-3-1023")),
                    PolicyAreaDefinition([10, 102, 1024], _("d-3-1024")),
                    PolicyAreaDefinition([10, 102, 1025], _("d-3-1025")),
                    PolicyAreaDefinition([10, 102, 1026], _("d-3-1026")),
                    PolicyAreaDefinition([10, 102, 1027], _("d-3-1027")),
                    PolicyAreaDefinition([10, 102, 1028], _("d-3-1028")),
                ]),
                PolicyAreaDefinition([10, 103], _("d-2-103"), [
                    PolicyAreaDefinition([10, 103, 1031], _("d-3-1031")),
                    PolicyAreaDefinition([10, 103, 1032], _("d-3-1032")),
                    PolicyAreaDefinition([10, 103, 1033], _("d-3-1033")),
                    PolicyAreaDefinition([10, 103, 1034], _("d-3-1034")),
                    PolicyAreaDefinition([10, 103, 1035], _("d-3-1035")),
                    PolicyAreaDefinition([10, 103, 1036], _("d-3-1036")),
                    PolicyAreaDefinition([10, 103, 1037], _("d-3-1037")),
                    PolicyAreaDefinition([10, 103, 1038], _("d-3-1038")),
                ]),
            ]),
            PolicyAreaDefinition([11], _("d-1-11"), [
                PolicyAreaDefinition([11, 111], _("d-2-111")),
                PolicyAreaDefinition([11, 112], _("d-2-112")),
                PolicyAreaDefinition([11, 113], _("d-2-113")),
                PolicyAreaDefinition([11, 114], _("d-2-114"), [
                    PolicyAreaDefinition([11, 114, 1141], _("d-3-1141")),
                    PolicyAreaDefinition([11, 114, 1142], _("d-3-1142")),
                ]),
                PolicyAreaDefinition([11, 115], _("d-2-115")),
            ]),
            PolicyAreaDefinition([12], _("d-1-12"), [
                PolicyAreaDefinition([12, 121], _("d-2-121")),
                PolicyAreaDefinition([12, 122], _("d-2-122")),
                PolicyAreaDefinition([12, 123], _("d-2-123")),
                PolicyAreaDefinition([12, 124], _("d-2-124")),
                PolicyAreaDefinition([12, 125], _("d-2-125"), [
                    PolicyAreaDefinition([12, 125, 1251], _("d-3-1251")),
                    PolicyAreaDefinition([12, 125, 1252], _("d-3-1252")),
                    PolicyAreaDefinition([12, 125, 1253], _("d-3-1253")),
                    PolicyAreaDefinition([12, 125, 1254], _("d-3-1254")),
                ]),
            ])
        ])



from cgi import FieldStorage
from datetime import date
from decimal import Decimal
from io import BytesIO

from onegov.form import Form
from onegov.swissvotes.fields import PolicyAreaField
from onegov.swissvotes.fields import SwissvoteDatasetField
from onegov.swissvotes.models import ColumnMapper
from psycopg2.extras import NumericRange
from xlsxwriter.workbook import Workbook


class DummyPostData(dict):
    def getlist(self, key):
        v = self[key]
        if not isinstance(v, (list, tuple)):
            v = [v]
        return v


class Test_swisscotes_dataset_field:

    @staticmethod
    def get_form_and_field():
        form = Form()
        field = SwissvoteDatasetField()
        field = field.bind(form, 'dataset')
        return form, field

    @staticmethod
    def get_field_storage():
        field_storage = FieldStorage()
        field_storage.type = 'application/excel'
        field_storage.filename = 'test.xlsx'
        return field_storage

    def test_validators(self):
        form, field = self.get_form_and_field()
        assert field()
        assert len(field.validators) == 2

    def test_field_corrupt(self):

        form, field = self.get_form_and_field()

        field_storage = self.get_field_storage()
        field_storage.file = BytesIO(b'Test')
        field.process(DummyPostData({'dataset': field_storage}))

        assert not field.validate(form)
        assert "Not a valid XLSX file." in field.errors

    def test_missing_sheet_data(self):

        form, field = self.get_form_and_field()

        file = BytesIO()
        workbook = Workbook(file)
        workbook.add_worksheet('CITATION')
        workbook.close()
        file.seek(0)

        field_storage = self.get_field_storage()
        field_storage.file = file
        field.process(DummyPostData({'dataset': field_storage}))

        # It raises for the first sheet it cant find
        assert not field.validate(form)
        assert "Sheet DATA is missing." in field.errors

    def test_missing_sheet_citations(self):

        form, field = self.get_form_and_field()

        file = BytesIO()
        workbook = Workbook(file)
        workbook.add_worksheet('DATA')
        workbook.close()
        file.seek(0)

        field_storage = self.get_field_storage()
        field_storage.file = file
        field.process(DummyPostData({'dataset': field_storage}))

        # It raises for the first sheet it cant find
        assert not field.validate(form)
        assert "Sheet CITATION is missing." in field.errors

    def test_empty(self):

        # Empty XLSX/sheet
        form, field = self.get_form_and_field()

        file = BytesIO()
        workbook = Workbook(file)
        workbook.add_worksheet('DATA')
        workbook.add_worksheet('CITATION')
        workbook.close()
        file.seek(0)

        field_storage = self.get_field_storage()
        field_storage.file = file
        field.process(DummyPostData({'dataset': field_storage}))

        assert not field.validate(form)
        assert "No data." in field.errors

    def test_missing_columns(self):

        form, field = self.get_form_and_field()
        mapper = ColumnMapper()
        columns = [value for value in mapper.columns.values()
                   if value != 'anzahl']

        file = BytesIO()
        workbook = Workbook(file)
        worksheet = workbook.add_worksheet('DATA')
        workbook.add_worksheet('CITATION')

        worksheet.write_row(0, 0, columns)
        worksheet.write_row(1, 0, columns)
        workbook.close()
        workbook.close()
        file.seek(0)

        field_storage = self.get_field_storage()
        field_storage.file = file

        field.process(DummyPostData({'dataset': field_storage}))

        assert not field.validate(form)
        errors = [error.interpolate() for error in field.errors]

        assert 'Some columns are missing: anzahl.' in errors

    def test_types_and_missing_values(self):

        form, field = self.get_form_and_field()
        mapper = ColumnMapper()
        file = BytesIO()
        workbook = Workbook(file)
        worksheet = workbook.add_worksheet('DATA')
        workbook.add_worksheet('CITATION')
        worksheet.write_row(0, 0, mapper.columns.values())
        worksheet.write_row(1, 0, [
            '',  # anr / NUMERIC
            '',  # datum / DATE
            '',  # legislatur / INTEGER
            '',  # legisjahr / INT4RANGE
            '',  # titel / TEXT
        ])
        worksheet.write_row(2, 0, [
            None,  # anr / NUMERIC
            None,  # datum / DATE
            None,  # legislatur / INTEGER
            None,  # legisjahr / INT4RANGE
            None,  # titel / TEXT
        ])
        worksheet.write_row(3, 0, [
            'x',  # anr / NUMERIC
            'x',  # datum / DATE
            'x',  # legislatur / INTEGER
            'x',  # legisjahr / INT4RANGE
            'x',  # titel / TEXT
        ])
        worksheet.write_row(4, 0, [
            1,  # anr / NUMERIC
            1,  # datum / DATE
            1,  # legislatur / INTEGER
            1,  # legisjahr / INT4RANGE
            1,  # titel / TEXT
        ])
        worksheet.write_row(5, 0, [
            1.1,  # anr / NUMERIC
            1.1,  # datum / DATE
            1.1,  # legislatur / INTEGER
            1.1,  # legisjahr / INT4RANGE
            1.1,  # titel / TEXT
        ])
        worksheet.write_row(5, 0, [
            date(2018, 12, 12),  # anr / NUMERIC
            date(2018, 12, 12),  # datum / DATE
            date(2018, 12, 12),  # legislatur / INTEGER
            date(2018, 12, 12),  # legisjahr / INT4RANGE
            date(2018, 12, 12),  # titel / TEXT
        ])
        workbook.close()
        file.seek(0)

        field_storage = self.get_field_storage()
        field_storage.file = file

        field.process(DummyPostData({'dataset': field_storage}))

        assert not field.validate(form)
        error = [error.interpolate() for error in field.errors][0]

        assert "1:anr ∅" in error
        assert "1:datum ∅" in error
        assert "1:legislatur ∅" in error
        assert "1:legisjahr ∅" in error
        assert "1:titel_off_d ∅" in error
        assert "1:titel_off_f ∅" in error
        assert "1:titel_kurz_d ∅" in error
        assert "1:titel_kurz_f ∅" in error

        assert "1:anzahl ∅" in error
        assert "1:rechtsform ∅" in error

        assert "2:anr ∅" in error
        assert "2:datum ∅" in error
        assert "2:legislatur ∅" in error
        assert "2:legisjahr ∅" in error
        assert "2:titel_off_d ∅" in error
        assert "2:titel_off_f ∅" in error
        assert "2:titel_kurz_d ∅" in error
        assert "2:titel_kurz_f ∅" in error

        assert "3:anr 'x' ≠ numeric(8, 2)" in error
        assert "3:datum 'x' ≠ date" in error
        assert "3:legislatur 'x' ≠ integer" in error
        assert "3:legisjahr 'x' ≠ int4range" in error

        assert "4:legisjahr '1' ≠ int4range" in error

        assert "5:legisjahr '43446' ≠ int4range" in error

    def test_all_okay(self):
        form, field = self.get_form_and_field()
        mapper = ColumnMapper()
        file = BytesIO()
        workbook = Workbook(file)
        worksheet = workbook.add_worksheet('DATA')
        workbook.add_worksheet('CITATION')
        worksheet.write_row(0, 0, mapper.columns.values())
        worksheet.write_row(1, 0, [
            '100.1',  # anr / NUMERIC
            '1.2.2008',  # datum / DATE
            '1',  # legislatur / INTEGER
            '2004-2008',  # legisjahr / INT4RANGE
            'titel_kurz_d',  # short_title_de / TEXT
            'titel_kurz_f',  # short_title_fr / TEXT
            'titel_off_d',  # title_de / TEXT
            'titel_off_f',  # title_fr / TEXT
            'stichwort',  # stichwort / TEXT
            '2',  # anzahl / INTEGER
            '3',  # rechtsform
        ])
        worksheet.write_row(2, 0, [
            100.2,  # anr / NUMERIC
            date(2008, 2, 1),  # datum / DATE
            1,  # legislatur / INTEGER
            '2004-2008',  # legisjahr / INT4RANGE
            'titel_kurz_d',  # short_title_de / TEXT
            'titel_kurz_f',  # short_title_fr / TEXT
            'titel_off_d',  # title_de / TEXT
            'titel_off_f',  # title_fr / TEXT
            'stichwort',  # stichwort / TEXT
            2,  # anzahl / INTEGER
            3,  # rechtsform
        ])
        workbook.close()
        file.seek(0)

        field_storage = FieldStorage()
        field_storage.file = file
        field_storage.type = 'application/excel'
        field_storage.filename = 'test.xlsx'

        field.process(DummyPostData({'dataset': field_storage}))

        assert field.validate(form)
        assert not field.errors

        assert field.data[0].bfs_number == Decimal('100.10')
        assert field.data[0].date == date(2008, 2, 1)
        assert field.data[0].legislation_number == 1
        assert field.data[0].legislation_decade == NumericRange(2004, 2008)
        assert field.data[0].title_de == 'titel_off_d'
        assert field.data[0].title_fr == 'titel_off_f'
        assert field.data[0].short_title_de == 'titel_kurz_d'
        assert field.data[0].short_title_fr == 'titel_kurz_f'
        assert field.data[0].keyword == 'stichwort'
        assert field.data[0].votes_on_same_day == 2
        assert field.data[0]._legal_form == 3

        assert field.data[1].bfs_number == Decimal('100.20')
        assert field.data[1].date == date(2008, 2, 1)
        assert field.data[1].legislation_number == 1
        assert field.data[1].legislation_decade == NumericRange(2004, 2008)
        assert field.data[1].title_de == 'titel_off_d'
        assert field.data[1].title_fr == 'titel_off_f'
        assert field.data[1].short_title_de == 'titel_kurz_d'
        assert field.data[1].short_title_fr == 'titel_kurz_f'
        assert field.data[1].keyword == 'stichwort'
        assert field.data[1].votes_on_same_day == 2
        assert field.data[1]._legal_form == 3


def test_policy_area_field():
    form = Form()
    field = PolicyAreaField(choices=[])
    field = field.bind(form, 'policy_area')

    html = field()
    assert 'class="policy-selector"' in html
    assert 'data-no-matches-text="No results match"' in html
    assert 'data-placehoder-text="Select Some Options"' in html
    assert 'data-tree="[]"' in html
    assert 'multiple' in html

    field.tree = [
        {
            'label': 'A',
            'value': 'a',
            'children': [
                {
                    'label': 'A.1',
                    'value': 'a1',
                    'children': []
                },
                {
                    'label': 'A.2',
                    'value': 'a2',
                    'children': [
                        {
                            'label': 'A.2.1',
                            'value': 'a21',
                            'children': []
                        }
                    ]
                }
            ],
        },
        {
            'label': 'B',
            'value': 'b',
            'children': []
        }
    ]
    assert field.choices == [
        ('a', 'A'), ('a1', 'A.1'), ('a2', 'A.2'), ('a21', 'A.2.1'), ('b', 'B')
    ]

    field.process(DummyPostData({'policy_area': ['a2', 'b']}))
    assert field.data == ['a2', 'b']
    assert field.tree == [
        {
            'label': 'A',
            'value': 'a',
            'checked': False,
            'expanded': True,
            'children': [
                {
                    'label': 'A.1',
                    'value': 'a1',
                    'checked': False,
                    'expanded': False,
                    'children': [],
                },
                {
                    'label': 'A.2',
                    'value': 'a2',
                    'checked': True,
                    'expanded': False,
                    'children': [
                        {
                            'label': 'A.2.1',
                            'value': 'a21',
                            'checked': False,
                            'expanded': False,
                            'children': [],
                        }
                    ],
                }
            ],
        },
        {
            'label': 'B',
            'value': 'b',
            'checked': True,
            'expanded': False,
            'children': [],
        }
    ]
